/***** sequenceData.c *********************************************
 * Description: Collection of routines for reading and 
 * manipulating sequence data.
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Sun Jun  6 10:34:31 2004.
 *****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sequenceData.h"
#include "stringUtil.h"
#include "eprintf.h"

static int lastSequence = 0;
static char *line = NULL;

/* convertToACGT: convert nucleotide data to acgt alphabet.
 */
void convertToAcgt(Sequence *seq){
  replace(seq->seq,'r','g');
  replace(seq->seq,'y','t');
  replace(seq->seq,'m','a');
  replace(seq->seq,'k','g');
  replace(seq->seq,'s','g');
  replace(seq->seq,'w','a');
  replace(seq->seq,'h','a');
  replace(seq->seq,'b','g');
  replace(seq->seq,'v','g');
  replace(seq->seq,'d','g');
  replace(seq->seq,'n','g');
  replace(seq->seq,'u','t');

  replace(seq->seq,'R','G');
  replace(seq->seq,'Y','T');
  replace(seq->seq,'M','A');
  replace(seq->seq,'K','G');
  replace(seq->seq,'S','G');
  replace(seq->seq,'W','A');
  replace(seq->seq,'H','A');
  replace(seq->seq,'B','G');
  replace(seq->seq,'V','G');
  replace(seq->seq,'D','G');
  replace(seq->seq,'N','G');
  replace(seq->seq,'U','T');
}


/* everything that is not [acgtACGT] is flagged by a -1 */
int *getRestrictedDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)emalloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = -1;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */

  return dic;

}

/* getDnaDictionary: create DNA dictionary */
int *getDnaDictionary(int *dic){
  int i;

  if(dic == NULL)
    dic = (int *)malloc((DICSIZE+1)*sizeof(int));

  for(i=0; i<DICSIZE; i++)
    dic[i] = 0;

  dic['a'] = 0;   /* a */
  dic['c'] = 1;   /* c */
  dic['g'] = 2;   /* g */
  dic['t'] = 3;   /* t */
  dic['A'] = 0;   /* A */
  dic['C'] = 1;   /* C */
  dic['G'] = 2;   /* G */
  dic['T'] = 3;   /* T */
  dic['r'] = dic['g'];
  dic['R'] = dic['g'];
  dic['y'] = dic['t'];
  dic['Y'] = dic['t'];
  dic['m'] = dic['a'];
  dic['M'] = dic['a'];
  dic['k'] = dic['g'];
  dic['K'] = dic['g'];
  dic['s'] = dic['g'];
  dic['S'] = dic['g'];
  dic['w'] = dic['a'];
  dic['W'] = dic['a'];
  dic['h'] = dic['a'];
  dic['H'] = dic['a'];
  dic['b'] = dic['g'];
  dic['B'] = dic['g'];
  dic['v'] = dic['g'];
  dic['V'] = dic['g'];
  dic['d'] = dic['g'];
  dic['D'] = dic['g'];
  dic['n'] = dic['g'];
  dic['N'] = dic['g'];
  dic['u'] = dic['t'];
  dic['U'] = dic['t'];

  return dic;
}


/* reverse and complement a sequence */
Sequence *revcomp(Sequence *seq){
  int i,j,n;
  char c;
  Sequence *newSeq;

  newSeq = (Sequence *)emalloc(sizeof(Sequence));

  n = strlen(seq->seq);
  newSeq->seq = (char *) emalloc((n+1)*sizeof(char));

  newSeq->id = strdup2(seq->id);  
  j=0;
  for(i = n-1; i >= 0; i--){
    c = seq->seq[i];
    switch(c){
    case 'a':
      newSeq->seq[j++] ='t';
      break;
    case 'c':
      newSeq->seq[j++] ='g';
      break;
    case 'g':
      newSeq->seq[j++] ='c';
      break;
    case 't':
      newSeq->seq[j++] ='a';
      break;
    case 'A':
      newSeq->seq[j++] ='T';
      break;
    case 'C':
      newSeq->seq[j++] ='G';
      break;
    case 'G':
      newSeq->seq[j++] ='C';
      break;
    case 'T':
      newSeq->seq[j++] ='A';
      break;
    default:
      newSeq->seq[j++] =c;
      break;
    }
  }
  newSeq->seq[n]='\0';
  return newSeq; 
}

/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getPermanentNextSequence(FILE *fp){
  Sequence *sequence;
  int seqlen, seqi, i, l;
  int currentBuffer;

  if(lastSequence){
    return NULL;
  }
  if(line == NULL){
    line = (char *)emalloc((SEQLINE+2)*sizeof(char));
    line = fgets(line,SEQLINE,fp);
  }
  /* make a sequence object */
  sequence = (Sequence *)emalloc(sizeof(Sequence));
  /* allocate memory for sequence id */
  sequence->id = (char *)emalloc((strlen(line)+1)*sizeof(char));
  /* copy sequence id */
  strcpy(sequence->id,chomp(line));
  /* allocate memory for sequence string */
  sequence->seq = (char *)emalloc((SEQBUFFER+1)*sizeof(char));
  seqlen = 0;
  currentBuffer = SEQBUFFER;
  seqi = 0;
  while((line=fgets(line,SEQLINE,fp)) != NULL){
    if(strstr(line,">") != NULL){
      sequence->seq[seqi++] = '\0';
      sequence->seq = (char *) realloc(sequence->seq, seqi * sizeof(char));
      return sequence;
    }
    if(strlen(line)>SEQLINE){
      printf("error in getNextSequence: cannot deal with lines longer than %d bp.\n",SEQLINE);
      printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
      exit(2);
    }
    l = strlen(line);
   /* disregard the final carriage return */
    if(line[l-1] == '\n')
      l--;
    seqlen += l; 
    if(seqlen > currentBuffer){
      currentBuffer += SEQBUFFER;
      sequence->seq = (char *)erealloc(sequence->seq,currentBuffer);
    }
    for(i=0;i<l;i++){
      sequence->seq[seqi++] = line[i];
    }
    /* sequence->seq = strncat(sequence->seq,line,strlen(line)-1); */
  }
  sequence->seq[seqi++] = '\0';
  sequence->seq = (char *) realloc(sequence->seq,seqi*sizeof(char));
  lastSequence = 1;
  return sequence;
}

void resetSequenceReader(){
  line = NULL;
  lastSequence = 0;
}

/* Get next sequence from an open data stream in FASTA format; this stream may be the stdin */
Sequence *getNextSequence(FILE *fp){
  Sequence *sequence;
  int seqlen, seqi, i, l;
  int currentBuffer;

  if(lastSequence){
    return NULL;
  }

  if(line == NULL){
    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    line = fgets(line,SEQLINE,fp);
  }

  /* make a sequence object */
  sequence = (Sequence *) malloc(sizeof(Sequence));
  /* allocate memory for sequence id */
  sequence->id = (char *)malloc((strlen(line)+1)*sizeof(char));
  /* copy sequence id */
  strcpy(sequence->id,line);
  /* allocate memory for sequence string */
  sequence->seq = (char *)malloc((SEQBUFFER+1)*sizeof(char));

  seqlen = 0;
  currentBuffer = SEQBUFFER;
  seqi = 0;
  while((line=fgets(line,SEQLINE,fp)) != NULL){
    if(strstr(line,">") != NULL){
      sequence->seq = (char *) realloc(sequence->seq,strlen(sequence->seq)+1);
      return sequence;
    }
    if(strlen(line)>SEQLINE){
      printf("error in getNextSequence: cannot deal with lines longer than %d bp.\n",SEQLINE);
      printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
      exit(2);
    }
    l = strlen(line);
   /* disregard the final carriage return */
    if(line[l-1] == '\n')
      l--;
    seqlen += l; 
    if(seqlen > currentBuffer){
      currentBuffer += SEQBUFFER;
      sequence->seq = (char *)erealloc(sequence->seq,currentBuffer);
    }
    for(i=0;i<l;i++){
      sequence->seq[seqi++] = line[i];
    }
    /* sequence->seq = strncat(sequence->seq,line,strlen(line)-1); */
  }
  sequence->seq = (char *) realloc(sequence->seq,strlen(sequence->seq)+1);
  lastSequence = 1;
  return sequence;
}

/* read FASTA file and store data; return size of data set */
Sequence *readFasta(char *file, Sequence *data, int *size, int *residues){
    FILE *fp;
    int first, l, numentry, maxseqlen, seqlen;
    int *lengthptr;
    Sequence *od;

    line = (char *)malloc((SEQLINE+2)*sizeof(char));

    /* open input file NB: this has got to be an external file, otherwise the rewind fails */
    fp = fopen(file, "r");
    if(fp == NULL){
	fprintf(stderr,"error in readFasta: could not open %s\n.",file);
	exit(1);
    }
    
    /* count number of entries and determine length of longest sequence */
    numentry = 0;
    maxseqlen = 0;
    seqlen = 0;
    *residues = 0;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strlen(line)>SEQLINE){
	    printf("error in readFasta: cannot deal with lines longer than %d bp.\n",SEQLINE);
	    printf("  change the SEQLINE parameter in file sequenceData.h and recompile.\n");
	    exit(2);
	}
	if(strstr(line,">") != NULL){
	    numentry++;
	    if(seqlen > maxseqlen)
		maxseqlen = seqlen;
	    seqlen = 0;
	}else{
	    l = strlen(line) - 1;
	    seqlen += l; 
	    *residues += l;
	}
    }
    *size = numentry;
    /* allocate memory */
    data = (Sequence *)malloc(numentry*sizeof(Sequence));
    lengthptr = (int *)malloc(numentry*sizeof(int));
    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    od = data;
    rewind(fp);
    /* Determine the length of each input sequence */
    lengthptr--;
    while((line=fgets(line,SEQLINE+2,fp)) != NULL){
	if(strstr(line,">") != NULL){
	    lengthptr++;
	    *lengthptr=0;
	}else{
	    l = strlen(line) - 1;
	    *lengthptr += l;
	}
    }

    line = (char *)malloc((SEQLINE+2)*sizeof(char));
    rewind(fp);

    first = 1;
    lengthptr -= numentry;
    data--;
    while((line=fgets(line,SEQLINE,fp)) != NULL){
	line = chomp(line);
	if(strstr(line,">") != NULL){           /* found new entry */
	    if(first == 0){
		data->seq -= *lengthptr;        /* readjust sequence pointer position */
	    }else{
		first = 0;
	    }
	    lengthptr++;
	    data++;
	    data->seq = (char *)malloc(*lengthptr * sizeof(char));
	    data->id = (char *)malloc((strlen(line)+1)*sizeof(char));
	    strcpy(data->id,line);
	}else{
	    strcpy(data->seq,line);
	    data->seq += strlen(line);
	}
    }
    /* readjust last sequence pointer */
    data->seq -= *lengthptr;
    free(line);
    return od;
}

/* freeSequence: free the data structure Sequence */
Sequence *freeSequence(Sequence *seq){
  free(seq->id);
  free(seq->seq);
  free(seq);
  return seq;
}
